import math
import random
import string

from django.db import models
from django.utils.translation import gettext as _

from ..get_settings import extract_model_kwargs as ek


class RandomIDMixin(models.Model):
    """
    Adds an `id` field which is randomly created
    
    Options: {
        $create_id: {
            min_length [int]: Min length for the id,
            choices [str]: String from what should be chosen from to generate the id
        }
    }
    """
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    
    id = models.CharField(
        unique=True,
        primary_key=True,
        **ek(___common_name, "id", {
            "verbose_name": _("ID"),
            "help_text": _("An unique ID for the object"),
            "null": False,
            "blank": False,
            "editable": False,
            "max_length": 63
        }),
    )  # type: str
    
    @classmethod
    def _get_id_length(cls, choices: string) -> int:
        amount = cls.objects.all().count()
        
        # Round up, because that's the next length we need
        return math.ceil(
            math.log(max(1, amount), len(choices))
        )
    
    @classmethod
    def _create_id(cls, min_length: int, choices: str) -> str:
        """
        The id will be randomly generated. The length will be automatically adjusted, depending on how many elements are
        already in the database.
        
        SETTINGS:
            - $create_id":
                min_length: int - Minimum length of the id
                choices: string - Contains all the available choices from what should be chosen from
        
        :return: The id
        """
        # Preparing values
        length: int = cls._get_id_length(choices=choices)
        length = max(min_length, length)
        
        # Creating id
        return "".join(
            random.choices(
                choices, k=length
            )
        )
    
    def save(self, *args, **kwargs):
        if self.id is None or self.id == "":
            # Getting kwargs
            class_kwargs = ek(self.___common_name, "$create_id", {
                "min_length": 7,
                "choices": string.ascii_letters + string.digits,
            })
            
            self.id = self._create_id(class_kwargs["min_length"], class_kwargs["choices"])
        
        return super().save(*args, **kwargs)



from typing import *

from ..utils.settings import get_setting
from ..typings import *

__all__ = [
    "extract_kwargs_for_field",
    "get_default_kwargs",
    "get_kwargs",
    "extract_model_kwargs"
]

_COMMON_DEFAULT_KWARGS_UPDATE = {
    "max_length": 127,
}


def extract_kwargs_for_field(
        kwargs: Kwargs,
        field: str,
        update_from: Optional[Kwargs] = None,
) -> Kwargs:
    if update_from is None:
        update_from = {}
    else:
        update_from = update_from.copy()
    
    new_kwargs = kwargs.get(field, kwargs)
    update_from.update(new_kwargs)
    
    return update_from


def get_default_kwargs() -> Kwargs:
    """Returns the default kwargs used for all fields"""
    default_kwargs: Kwargs = get_setting("COMMON_DEFAULT_KWARGS", {})
    default_kwargs.update(_COMMON_DEFAULT_KWARGS_UPDATE)
    
    return default_kwargs


def get_kwargs(class_name: str) -> Kwargs:
    """Returns the specific kwargs for each field `class_name`"""
    default_kwargs = get_default_kwargs()
    class_kwargs = get_setting("COMMON_KWARGS", {})
    
    use_kwargs = class_kwargs.get(class_name, default_kwargs)
    
    return use_kwargs


def extract_model_kwargs(class_name: str, field: str, update_from: Optional[Kwargs] = None) -> Kwargs:
    kwargs: Kwargs = get_kwargs(class_name)
    
    return extract_kwargs_for_field(kwargs.get(
        field, get_default_kwargs()
    ), field, update_from)



from types import FunctionType
from typing import *

import htmlmin
from bs4 import BeautifulSoup, Tag

from .text import TextOptimizer
from ..constants import AddAttributesDict, HTMLOptimizerDefault, UnwrapDict
from ...typings import Kwargs
from ...utils import iteration


class TextHTMLOptimizer:
    """
    Optimizes a given string. Accepts text/html
    """
    
    # Dictionary containing as key a list with strings which tags should be unwrapped, the value is a list
    # containing strings with tags which triggers the unwrap action
    
    @staticmethod
    def minify_html(html: str, opts: Optional[Kwargs] = None) -> str:
        if opts is None:
            opts = HTMLOptimizerDefault.minify_opts
        
        return htmlmin.minify(html, **opts)
    
    @staticmethod
    def _change_text(soup: BeautifulSoup, func: FunctionType, *args, **kwargs) -> BeautifulSoup:
        for element in soup.find_all():  # type: Tag
            if len(element.find_all()) > 0:
                continue
            
            new_text = func(element.get_text(), *args, **kwargs)
            element.string = new_text
        
        return soup
    
    @classmethod
    def html_remove_redundant_space(cls, html: str, *args, **kwargs) -> str:
        soup = BeautifulSoup(html, "html.parser")
        cls._change_text(soup, TextOptimizer.remove_redundant_space, *args, **kwargs)
        return str(soup)
    
    @classmethod
    def html_space_after_text(cls, html: str, *args, **kwargs) -> str:
        soup = BeautifulSoup(html, "html.parser")
        cls._change_text(soup, TextOptimizer.space_after_text, *args, **kwargs)
        return str(soup)
    
    @classmethod
    def html_space_before_text(cls, html: str, *args, **kwargs) -> str:
        soup = BeautifulSoup(html, "html.parser")
        cls._change_text(soup, TextOptimizer.space_before_text, *args, **kwargs)
        return str(soup)
    
    @staticmethod
    def html_unwrap(html: str, unwrap: Optional[UnwrapDict] = None) -> str:
        """
        Unwraps html tags.
        Example:
            input: <p><img src="image.jpg" /></p>
            self.unwrap: {
                "p": "img"
            }
            output: <img src="image.jpg" />
        """
        
        if unwrap is None:
            unwrap = HTMLOptimizerDefault.unwrap
        
        soup = BeautifulSoup(html, "html.parser")
        
        for unwrap, trigger in iteration.ensure_dict(unwrap, str, str):
            for element in soup.find_all(unwrap):
                if element.find_all(trigger, recursive=False):
                    element.unwrap()
        
        return str(soup)
    
    @staticmethod
    def html_add_attributes_to_tags(html: str, add_attributes: Optional[AddAttributesDict] = None) -> str:
        """Adds attributes to html tags"""
        
        if add_attributes is None:
            add_attributes = HTMLOptimizerDefault.add_attributes
        
        soup = BeautifulSoup(html, "html.parser")
        
        for tag, attributes in iteration.ensure_dict(add_attributes, str, dict):
            for element in soup.find_all(tag):
                for key, value in iteration.ensure_dict(attributes, str, str):
                    element[key] = value
        
        return str(soup)



import re
from dataclasses import dataclass

from .base import BaseHandlerMixin
from ..constants import HandleOn, TextOptimizerDefault

__all__ = [
    "RegexHandler", "WhiteSpaceStripHandler", "TextOptimizerHandler"
]

from ..optimizers.text import TextOptimizer


class RegexHandler(BaseHandlerMixin):
    PATTERN = ""
    REPLACEMENT = " "
    
    @staticmethod
    def HANDLE_ON():
        return {HandleOn.CREATION, HandleOn.SAVE}
    
    def handle(self, value: str) -> str:
        return re.sub(self.__class__.PATTERN, self.__class__.REPLACEMENT, value)


class WhiteSpaceStripHandler(RegexHandler):
    PATTERN = r"\s+"
    REPLACEMENT = " "
    
    def handle(self, value: str) -> str:
        return super().handle(value).lstrip().rstrip()


@dataclass
class TextOptimizerHandler(BaseHandlerMixin):
    space_after: str = TextOptimizerDefault.space_after
    space_before: str = TextOptimizerDefault.space_before
    no_space_before: str = TextOptimizerDefault.no_space_before
    no_space_after: str = TextOptimizerDefault.no_space_after
    ignore_for_digits: str = TextOptimizerDefault.ignore_for_digits
    
    @staticmethod
    def HANDLE_ON():
        return {HandleOn.CREATION, HandleOn.SAVE}
    
    def handle(self, value: str) -> str:
        return TextOptimizer.space_before_text(
            TextOptimizer.space_after_text(
                TextOptimizer.remove_redundant_space(
                    value,
                    no_space_after=self.no_space_after,
                    no_space_before=self.no_space_before
                ),
                space_after=self.space_after,
                ignore_for_digits=self.ignore_for_digits,
            ),
            self.space_before
        )



from typing import *

from django_hint import RequestType


def get_ip(request: RequestType) -> str:
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def get_previous_site(request: RequestType) -> Optional[str]:
    return request.META.get("HTTP_REFERER")



from abc import ABC, abstractmethod
from typing import *

from ..constants import HandleOn

__all__ = [
    "BaseHandlerMixin"
]


class BaseHandlerMixin(ABC):
    """The BaseHandlerMixin for all handlers. All methods should be static, except your handler needs special
    passed parameters. """
    
    @staticmethod
    @abstractmethod
    def HANDLE_ON() -> Union[Iterable[str], str]:
        return HandleOn.SAVE
    
    @abstractmethod
    def handle(self, value):
        raise NotImplementedError(f"{__qualname__} is not implemented.")



from datetime import datetime
from typing import Optional
from urllib.parse import urlencode

from . import constants

# Example
"""
https://www.google.com/calendar/render
?action=TEMPLATE&text=Your+Event+Name
&dates=20140127T224000Z/20140320T221500Z
&details=For+details,+link+here:+http://www.example.com&location=Waldorf+Astoria,+301+Park+Ave+,+New+York,+NY+10022
&sf=true
&output=xml

http://www.google.com/calendar/render?
action=TEMPLATE
&text=[event-title]
&dates=[start-custom format='Ymd\\THi00\\Z']/[end-custom format='Ymd\\THi00\\Z']
&details=[description]
&location=[location]
&trp=false
"""

__all__ = [
    "create_gce_url"
]


def create_gce_url(
        text: str,
        start_date: datetime,
        end_date: Optional[datetime] = None,
        url: str = constants.URL,
        action: str = "TEMPLATE",
        date_format: str = "%Y%m%dT%H%M00",
        trp: bool = True,
        **kwargs
) -> str:
    """
    Creates a link to add an event to Google Calendar.
    Documentation: http://useroffline.blogspot.com/2009/06/making-google-calendar-link.html
    Stackoverflow answer: https://stackoverflow.com/a/23495015/9878135
    
    :param text: Parameter text
    :param start_date: Parameter start date, will be parsed with format date_format
    :param end_date: Parameter end_date, will be parsed with format date_format
    :param url: The url (works as prefix)
    :param action: Parameter action (default 'TEMPLATE')
    :param date_format: Date format that should be used to parse start_date and end_date
    :param trp: Parameter trp
    :param kwargs: Extra parameters that should be added to the link
    :return: str, ready-to-use link
    """
    
    date_string = start_date.strftime(date_format)
    
    if end_date:
        date_string += f"/{end_date.strftime(date_format)}"
    
    parsed = urlencode({
        "action": action,
        "text": text,
        "dates": date_string,
        "trp": trp,
        **kwargs
    })
    
    return f"{url}?{parsed}"



from typing import *

from django_hint import RequestType

from .helpers.queryset import CustomQuerySet, CustomQuerySetManager


class CustomQuerySetMixin:
    objects = CustomQuerySetManager()
    
    class QuerySet(CustomQuerySet):
        def accessible(self, request: Optional[RequestType] = None):
            return self
        
        def visible(self, request: Optional[RequestType] = None):
            return self.accessible(request)



from django.template.defaulttags import register
from django_hint import QueryType


@register.filter
def slice_modulo(qs: QueryType, mod: int) -> QueryType:
    """Slices queryset to a specific modulo"""
    found = len(qs)
    
    for i in range(found, 1, -1):
        if i % mod == 0:
            found = i
            break
    
    return qs[:found]



from django.db import models


class CustomQuerySetManager(models.Manager):
    """A re-usable Manager to access a custom QuerySet"""
    
    def __getattr__(self, attr, *args):
        try:
            return getattr(self.__class__, attr, *args)
        except AttributeError:
            # don't delegate internal methods to the queryset
            if attr.startswith('__') and attr.endswith('__'):
                raise
            return getattr(self.get_query_set(), attr, *args)
    
    def get_queryset(self):
        return self.model.QuerySet(self.model)
    
    def get_querySet(self):
        return self.get_queryset()
    
    def get_query_set(self):
        return self.get_queryset()


class CustomQuerySet(models.QuerySet):
    def __getattr__(self, attr, *args):
        try:
            return getattr(super().__class__, attr, *args)
        except AttributeError:
            return getattr(self.model.QuerySet, attr, *args)



from django_common_utils.libraries.handlers import HandlerDefinitionType, HandlerMixin
from django_common_utils.libraries.handlers.mixins import WhiteSpaceStripHandler
from django_common_utils.libraries.models import TitleMixin


class Article(HandlerMixin, TitleMixin):
    @staticmethod
    def handlers() -> HandlerDefinitionType:
        return {
            "title": WhiteSpaceStripHandler
        }



from dataclasses import dataclass

from .typings import *
from ..typings import ModelInstance

__all__ = [
    "ApplyHandler"
]


@dataclass
class ApplyHandler:
    handlers: HandlerDefinitionType
    instance: ModelInstance
    
    def handle(self) -> AppliedHandlersType:
        fields: ApplyHandlerDefinitionType = {}
        
        for field, Handler in self.handlers.items():
            # Get current value, either from the instance or from previously handled handlers
            current_value = fields.get(field, getattr(self.instance, field))
            # Get new value
            new_value = Handler.handle(current_value)
            # Safe new value
            fields[field] = new_value
        
        return fields



from django.shortcuts import render
from django_hint import RequestType

from development.article.models import Article

def test(request: RequestType):
    return render(request, "test.html", {
        "list": ["Verstehen", "Sie", "Spass"],
        "article": Article
    })



from typing import *

from .mixins.base import BaseHandlerMixin
from ..utils.typings import *

__all__ = [
    "HandlerDefinitionType", "HandlerInstance", "AppliedHandlersType", "ApplyHandlerDefinitionType"
]

HandlerInstance = Type[BaseHandlerMixin]
# Dict[field_name(s), handler(s)]
HandlerDefinitionType = EnsureIterationDictType[str, HandlerInstance]
# Dict[field_name, value]
AppliedHandlersType = Dict[str, Any]
# Dict[field_name, handler]
ApplyHandlerDefinitionType = Dict[str, HandlerInstance]



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from .libraries.handlers import HandleOn, HandlerMixin


# noinspection PyProtectedMember
@receiver(post_save)
def handler_save(sender, instance, created: bool, *args, **kwargs) -> None:
    if issubclass(sender, HandlerMixin):
        if created:
            instance._apply_handlers(HandleOn.CREATION)
        else:
            instance._apply_handlers(HandleOn.SAVE)


# noinspection PyProtectedMember
@receiver(post_delete)
def handler_delete(sender, instance, *args, **kwargs) -> None:
    if issubclass(sender, HandlerMixin):
        instance._apply_handlers(HandleOn.DELETION)



from ..admin import AdminFieldsetMixin


class AuthorAdminFieldsetMixin(AdminFieldsetMixin):
    def get_readonly_fields(self, obj=None, **_):
        if obj:
            return "authors"
    
    def get_mixin_fields(self, **_) -> dict:
        return {
            "default": ["authors", ]
        }


class AutomaticUserAssociationCreationFieldsetMixin(AdminFieldsetMixin):
    def get_readonly_fields(self, obj=None, **_):
        if obj:
            return "added_by"
    
    def get_mixin_fields(self, **_):
        return {
            "created": ["added_by"]
        }



from typing import *

T = TypeVar("T")
S = TypeVar("S")

__all__ = [
    "EnsureIterationType", "EnsureIterationDictType"
]

EnsureIterationType = Tuple[Iterable[T], T]
EnsureIterationDictType = Dict[EnsureIterationType[T], EnsureIterationType[S]]



from django.contrib import admin


class AutomaticUserAssociationAdminMixin(admin.ModelAdmin):
    def save_model(self, request, obj, *args, **kwargs):
        if not obj:
            obj.added_by = request.user
        return super().save_model(request, obj, *args, **kwargs)



from django import forms
from django.forms import Field

__all__ = [
    "UpdateDefaultsForm"
]


class UpdateDefaultsForm(forms.ModelForm):
    """ModelForm to automatically set the initial values from the instance."""
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_instance_values()
    
    def set_instance_values(self):
        if self.instance:
            name: str
            field: Field
            for name, field in self.fields.items():
                self.initial[name] = getattr(self.instance, name, field.initial)



try:
    from django.contrib.sites.models import Site
except RuntimeError:
    pass

try:
    current_site = Site.objects.get_current()  # type: Site
except:
    class Other:
        @property
        def domain(self):
            return "example.com"
        
        @property
        def name(self):
            return "example.com"
    
    
    current_site = Other()



import math
import operator
from functools import reduce

from django.template.defaulttags import register

from ..libraries.typings import Number
from ..libraries.utils.templatetags_factory import iter_module


@register.simple_tag
def add(*args) -> Number:
    return reduce(operator.__add__, args)


@register.simple_tag
def subtract(*args) -> Number:
    return reduce(operator.__sub__, args)


@register.simple_tag
def multiply(*args) -> Number:
    return reduce(operator.__mul__, args)


@register.simple_tag
def divide(*args) -> Number:
    return reduce(operator.__floordiv__, args)


for name, function in iter_module(math):
    register.simple_tag(
        function,
        name=name
    )



from pathlib import Path
from typing import *

from django.conf import settings
from django.utils.safestring import mark_safe
from PIL import Image


def generate_image(
        url: str,
        alt: str = "",
        background_color: str = "rgba(0, 0, 0, 0)",
        title: Optional[str] = None
) -> str:
    """
    Generates an <img /> tag with path as picture.

    :param url: Pah to the image
    :type url: str

    :param alt: Alt property
    :type alt: str

    :param background_color: The background color
    :type background_color: str

    :param title: Title property
    :type title: str

    :return: <img /> tag
    """
    
    absolute_path = Path(settings.BASE_DIR).joinpath(url[1:])
    extra = {}
    
    if title:
        alt = f"{title} - {alt}"
    
    try:
        with Image.open(str(absolute_path)) as image:
            extra["width"], extra["height"] = image.size
    except (FileNotFoundError, PermissionError):
        pass
    
    attributes = " ".join([f"{key}={value}" for key, value in libraries.items()])
    
    return mark_safe(
        f'<img src="{url}" alt="{alt}" loading="lazy" style="background-color:{background_color}" {attributes} />'
    )



from datetime import datetime

from django.db import models
from django.utils.translation import gettext as _

from ..get_settings import extract_model_kwargs as ek


class CreationDateMixin(models.Model):
    """Adds an `created_at` field, which stores date and time, when the object was created"""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    
    created_at = models.DateTimeField(
        **ek(___common_name, "created_at", {
            "verbose_name": _("Creation date"),
            "help_text": _("Date and time when this was created"),
            "editable": False,
        })
    )  # type: datetime
    
    def save(self, *args, **kwargs):
        if self.created_at is None:
            self.created_at = datetime.now()
        
        return super().save(*args, **kwargs)


class EditCreationDateMixin(CreationDateMixin):
    """Adds an `edited_at` field, which stores date and time, when this object was modified last. The field is empty,
    if the object hasn't been edited since creation."""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    
    edited_at = models.DateTimeField(
        **ek(___common_name, "edited_at", {
            "verbose_name": _("Edit date"),
            "help_text": _("Last date and time when this was edited"),
            "editable": False,
            "blank": True,
            "null": True,
        })
    )  # type: datetime
    
    def save(self, *args, **kwargs):
        if self.created_at is not None:
            self.edited_at = datetime.now()
        
        return super().save(*args, **kwargs)



from django.contrib import admin

from django_common_utils.libraries.fieldsets import DefaultAdminMixin
from django_common_utils.libraries.fieldsets.mixins import TitleAdminFieldsetMixin
from .models import Article


class ArticleAdmin(DefaultAdminMixin):
    mixins = [
        TitleAdminFieldsetMixin,
    ]
    fieldset_fields = {
        "default": ["clicks", "...!"]
    }


admin.site.register(Article, ArticleAdmin)



from django.db import models
from django.utils.translation import gettext as _

from ...utils.text import create_short
from ..get_settings import extract_model_kwargs as ek


class TitleMixin(models.Model):
    """Adds a `title` field"""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    ___COMMON_SHORT_TITLE_LENGTH: int = 50  # Length for the `short_title`
    ___COMMON_TITLE_CHANGE_THRESHOLD: int = 24
    
    title = models.CharField(
        **ek(___common_name, "title", {
            "verbose_name": _("Title"),
            "max_length": 60,
        })
    )  # type: str
    
    def __str__(self):
        return self.short_title
    
    @property
    def short_title(self) -> str:
        return create_short(self.title, self.___COMMON_SHORT_TITLE_LENGTH)


class DescriptionMixin(models.Model):
    """Adds a `description` field"""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    ___COMMON_SHORT_DESCRIPTION_LENGTH: int = 85  # Length for the `short_title`
    
    description = models.CharField(
        **ek(___common_name, "description", {
            "verbose_name": _("Description"),
            "max_length": 120,
            "blank": True,
            "null": True
        })
    )  # type: str
    
    @property
    def get_description(self) -> str:
        return self.description if self.description is not None else ""
    
    @property
    def short_description(self):
        return create_short(self.description, self.___COMMON_SHORT_DESCRIPTION_LENGTH) \
            if self.description is not None else ""



from datetime import datetime, timedelta
from typing import Optional

from ..admin import AdminFieldsetMixin


class TitleAdminFieldsetMixin(AdminFieldsetMixin):
    def get_readonly_fields(self, obj=None, **_) -> Optional[str]:
        # If object has a pub_date, check if that date is within the allowed time range from at_date.
        # If the date is older than the allowed range, make title read-only.
        # alternative_title will be turned read-only, if pub_date is not in the allowed range and alternative_title
        # is already set.
        if obj:
            if (date := hasattr(obj, "pub_date")) or (date := hasattr(obj, "edited_at")):
                now = datetime.now()
                # Check time range
                if now - timedelta(hours=obj.___COMMON_TITLE_CHANGE_THRESHOLD) > date:
                    return "title"
            else:
                return "title"
    
    def get_mixin_fields(self, **_) -> dict:
        return {
            "default": ["title"],
        }


class DescriptionAdminFieldsetMixin(AdminFieldsetMixin):
    def get_mixin_fields(self, **_) -> dict:
        return {
            "default": ["description", ]
        }


class TitleAdminFieldsetMixin(AdminFieldsetMixin):
    def get_mixin_fields(self, **_):
        return {
            "default": ["title"]
        }
    
    def get_readonly_fields(self, obj=None, **_):
        if obj:
            return "title"



import re

from ...handlers import TextOptimizerDefault


class TextOptimizer:
    """
    Optimizes a given string. Accepts text/plain.
    """
    
    @staticmethod
    def remove_redundant_space(
            text: str,
            no_space_after: str = TextOptimizerDefault.no_space_after,
            no_space_before: str = TextOptimizerDefault.no_space_before
    ) -> str:
        return re.sub(
            rf"([{no_space_after}])\s+",
            r"\1",
            re.sub(
                rf'\s+([{no_space_before}])',
                r"\1",
                re.sub(r"\s\s+", " ", text)
            )
        )
    
    @staticmethod
    def space_after_text(
            text: str,
            space_after: str = TextOptimizerDefault.space_after,
            ignore_for_digits: str = TextOptimizerDefault.ignore_for_digits
    ) -> str:
        return re.sub(
            rf"(?<=[{space_after}])(?=[^\s])(?<![\d{ignore_for_digits}](?=\d))", " ", text
        )
    
    @staticmethod
    def space_before_text(
            text: str,
            space_before: str = TextOptimizerDefault.space_after
    ) -> str:
        return re.sub(
            rf"(?<=[^\s])(?=[{space_before}])", " ", text
        )



import re
from typing import *

from django.utils.html import strip_tags
from django.utils.translation import gettext as _

from ..typings import ModelInstance


def create_short(text: str, max_length: int, prefix: str = "...") -> str:
    return f"{text[:max_length]}{prefix}" if len(text) > max_length else text


def model_verbose(model: ModelInstance) -> str:
    return model._meta.verbose_name


def model_verbose_plural(model: ModelInstance) -> str:
    return model._meta.verbose_name_plural


def camelcase_to_underscore(value: str) -> str:
    x: str
    return "_".join(
        [x.lower() for x in re.findall("[A-Z][^A-Z]*", value)]
    )


def listify(x: List[str], /) -> str:
    length = len(x)
    
    if length == 0:
        return ""
    elif length == 1:
        return x[0]
    return _("{} and {}").format(
        ", ".join([str(value) for value in x[:-1]]),
        x[-1]
    )


def field_verbose(model: ModelInstance, field: str) -> str:
    return model._meta.get_field(field).verbose_name


def field_verbose_plural(model: ModelInstance, field: str) -> str:
    return model._meta.get_field(field).verbose_name_plural


def textify(html: str) -> str:
    """Textifies an html input"""
    # Remove html tags and continuous whitespaces
    text_only = re.sub("[ \t]+", " ", strip_tags(html))
    # Strip single spaces in the beginning of each line
    return text_only.replace('\n ', '\n').strip()



from dataclasses import dataclass
from typing import *

from .base import BaseHandlerMixin
from ..constants import AddAttributesDict, HandleOn, HTMLOptimizerDefault, TextOptimizerDefault, UnwrapDict
from ..optimizers.html import TextHTMLOptimizer
from ...typings import *


@dataclass
class HTMLTextOptimizer(BaseHandlerMixin):
    space_after: str = TextOptimizerDefault.space_after
    space_before: str = TextOptimizerDefault.space_before
    no_space_before: str = TextOptimizerDefault.no_space_before
    no_space_after: str = TextOptimizerDefault.no_space_after
    ignore_for_digits: str = TextOptimizerDefault.ignore_for_digits
    
    def __init__(
            self,
            unwrap: Optional[UnwrapDict] = None,
            add_attributes: Optional[AddAttributesDict] = None,
            minify_opts: Optional[Kwargs] = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.unwrap = unwrap if unwrap is not None else HTMLOptimizerDefault.unwrap
        self.add_attributes = add_attributes if add_attributes is not None else HTMLOptimizerDefault.add_attributes
        self.minify_opts = minify_opts if minify_opts is not None else HTMLOptimizerDefault.minify_opts
    
    @staticmethod
    def HANDLE_ON():
        return {HandleOn.CREATION, HandleOn.SAVE}
    
    def handle(self, value: str):
        return TextHTMLOptimizer.html_remove_redundant_space(
            TextHTMLOptimizer.html_space_after_text(
                TextHTMLOptimizer.html_space_before_text(
                    TextHTMLOptimizer.html_add_attributes_to_tags(
                        TextHTMLOptimizer.html_unwrap(
                            TextHTMLOptimizer.minify_html(value, opts=self.minify_opts),
                            unwrap=self.unwrap
                        ),
                        add_attributes=self.add_attributes
                    ),
                    space_before=self.space_before
                ),
                space_after=self.space_after,
                ignore_for_digits=self.ignore_for_digits
            ),
            no_space_after=self.no_space_after,
            no_space_before=self.no_space_before,
        )



import collections
import logging
from abc import abstractmethod

from django.core.exceptions import FieldDoesNotExist
from django.db.models.options import Options

from .handlers import ApplyHandler
from .mixins.base import BaseHandlerMixin
from .typings import *
from .typings import ApplyHandlerDefinitionType
from ..utils import iteration

__all__ = [
    "HandlerMixin"
]


def is_valid_handler(value) -> bool:
    return issubclass(value, BaseHandlerMixin)


class HandlerMixin:
    @staticmethod
    @abstractmethod
    def handlers() -> HandlerDefinitionType:
        return {}  # No handlers
    
    def _get_true_handlers(self, action: str) -> ApplyHandlerDefinitionType:
        """Gets all valid handlers and fields."""
        true_handlers = collections.defaultdict(list)
        
        meta: Options = self._meta
        
        field: str
        handler: HandlerInstance
        for field, handler in iteration.ensure_dict(
                self.__class__.handlers(),  # Ensures that handlers are static
                str,  # Field names are strings
                lambda instance: is_valid_handler(instance)  # Use function to determine validness of handlers
        ):
            # Check field
            try:
                meta.get_field(field)
            except FieldDoesNotExist:
                logging.warning(f'Field "{field}" does not exist on instance {self}, skipping it.')
                continue
            
            # Check handler
            # Check if actual handler
            if not is_valid_handler(handler):
                logging.warning(f'Handler {handler} is not a valid handler, skipping it.')
                continue
            # Check HANDLE_ON
            if action not in handler.HANDLE_ON():
                continue
            
            true_handlers[field].append(handler)
        
        return dict(true_handlers)
    
    def _apply_handlers(self, action: str) -> None:
        """Applies all specified handlers onto this instance. Fields wil be overwritten!"""
        applier = ApplyHandler(self, self._get_true_handlers(), action)
        
        for field, value in applier.handle():
            setattr(self, field, value)



from collections import Iterable
from typing import List, Optional, Union


def flat_list(lst: list) -> list:
    flt_lst = []
    
    for value in lst:
        if isinstance(value, list):
            flt_lst.extend(flat_list(value))
        else:
            flt_lst.append(value)
    
    return flt_lst


class UniqueSortedDict(dict):
    def __getitem__(self, item):
        # sort dict
        sorted_dict = {k: v for k, v in sorted(self.items(), key=lambda x: x[0])}
        return sorted_dict[item]
    
    def __setitem__(self, key, value):
        if key not in self:
            dict.__setitem__(self, key, value)
        else:
            raise KeyError("Key already in use!")
    
    def __iter__(self):
        for key, value in self.items():
            yield value


class FieldsetList:
    INSERT_STRING = "..."
    FORCE_INSERT_CHAR = "!"
    FORCE_INSERT_BEFORE = f"{FORCE_INSERT_CHAR}{INSERT_STRING}"
    FORCE_INSERT_AFTER = f"{INSERT_STRING}{FORCE_INSERT_CHAR}"
    
    def __init__(
            self,
            fields: list,
            name: str,
            appearance: Optional[str] = None,
            classes: list = [],
            description: str = "",
            order: int = 0,
    ):
        self.fields = fields
        self.appearance = appearance
        self.classes = classes
        self.name = name
        self.description = description
        self.order = order
    
    def __str__(self):
        return f"{self.appearance}: {str(self.fields)}"
    
    def __repr__(self):
        return self.__str__()
    
    def __getitem__(self, item: int):
        if isinstance(item, int):
            return self.fields[item]
        raise KeyError("You can only slice using int!")
    
    def __add_field(self, value: Union[str, Iterable], index: Optional[int] = None):
        if type(value) is str:
            value: List[str] = [value]
        
        for val in value:
            if val not in self.fields:
                if index:
                    self.fields[index] = val
                else:
                    self.fields.append(val)
    
    def __setitem__(self, key: int, value):
        if isinstance(key, int):
            self.__add_field(value, key)
        raise KeyError("You can only set an value using an int as the key!")
    
    def __insert_fields(self, adds: Union[str, Iterable], insert: str):
        if isinstance(adds, str):
            adds = {adds}
        
        fields: list
        if insert in (fields := self.fields):
            index = fields.index(insert)
            # Get not already declared fields
            adds = [field for field in adds if field not in fields]
            
            # PERFORMANCE: Using `[index:index]` method of fields and adding `len` to index to avoid incrementing
            # index multiple times
            fields[index:index] = adds
            
            # If FORCE_INSERT_BEFORE is used, move it to original position
            if insert == self.FORCE_INSERT_BEFORE:
                # Get index of FORCE_INSERT_BEFORE
                insert_index = index + len(adds)
                
                # Copy insert to old position
                fields.insert(index, fields[insert_index])
                insert_index += 1  # Update index
                
                # Delete copied insert on new position
                del fields[insert_index]
            elif insert != self.FORCE_INSERT_AFTER:  # No need to check FORCE_INSERT_BEFORE again
                index += len(adds)  # Update index of INSERT_STRING
                del fields[index]
            
            # Update fields
            self.fields = fields
            
            return self
        raise KeyError(f'"{insert}" was not found in fields!')
    
    def append_fields(self, adds: Union[str, Iterable]):
        """
            Appends given fields to end of current fields.
        :param adds: If str, will be appended.
        If Iterable, will be extended.
        :return: self
        """
        self.__add_field(adds)
        
        return self
    
    def add_fields(self, fields: Union[str, Iterable], insert_prefix: str = ""):
        """
            Adds fields.
        :param fields: List, containing str as fields to be added
        :param insert_prefix: Prefix for the INSERT_STRING
        :return: self
        """
        
        def get_insert():
            own_fields = self.fields
            checking = [
                f"{insert_prefix}{self.INSERT_STRING}",
                self.INSERT_STRING,
                self.FORCE_INSERT_BEFORE,
                self.FORCE_INSERT_AFTER
            ]
            founds = [field for field in checking if field in own_fields]
            
            if founds:
                return founds[0]
            return
        
        if insert := get_insert():
            self.__insert_fields(fields, insert)
        else:
            self.append_fields(fields)
        
        return self
    
    def __add__(self, other: Union[str, Iterable]):
        self.append_fields(other)
        return self
    
    def __sub__(self, other: str):
        """
            Removes a field from fields.
        :param other: The field to be removed
        :return: self
        """
        self.fields.remove(other)
        return self
    
    def integrate(self, value: "FieldsetList"):
        """
            Extends fields and classes of the other FieldsetList.
        :param value: The other FieldsetList (This fields and classes will be extended to self's fields and classes)
        :return: self
        """
        self.__add_field(value.fields)
        self.classes.extend(value.classes)
        
        return self
    
    def clear_fields_copy(self) -> list:
        """
            Creates a cleared copy of fields.
        :return: The copied, cleared fields
        """
        fields = self.fields
        
        for field in fields:
            if self.INSERT_STRING in field:
                fields.remove(field)
        
        return fields
    
    def render(self) -> list:
        return [
            self.appearance, {
                "fields": self.fields,
                "classes": self.classes,
                "description": self.description
            }
        ]
    
    @property
    def flatten_fields(self):
        return flat_list(self.clear_fields_copy())


class Sections:
    def __init__(self, sets: list, move_fields: bool = False):
        self._fields = UniqueSortedDict()
        self.move_fields = move_fields
        
        for fieldset in sets:  # type: FieldsetList
            self.__add_fieldsetlist(fieldset)
    
    def get(self, value) -> Union[FieldsetList, list, None]:
        try:
            return self[value]
        except (KeyError, StopIteration):
            return None
    
    def __getitem__(self, item: Union[int, slice, str, list]) -> Union[FieldsetList, list]:
        """
            Returns a FieldsetList based on item.
        :param item: int, slice, str or list.
        If item is int, the first FieldsetList whose order matches will be returned.
        If item is slice, a list with FieldsetList with their orders in range will be returned.
        If item is str, the first FieldsetList whose name matches will be returned.
        If item ist list, the first FieldsetList whose fields matches (will be checked by converting both (item and
        the fields of the current FieldsetList) to a set).
        :return: FieldsetList or list containing FieldsetLists
        """
        field: FieldsetList
        
        # If item is int, return whose order matches with item
        if isinstance(item, int):
            return self._fields[item]
        
        # If item is slice, return whose order is in range with item
        elif isinstance(item, slice):
            order_range = range(item.start, item.stop, item.step)
            return [field for field in self._fields if field.order in order_range]
        
        # If item is str, return whose appearance matches with item
        elif isinstance(item, str):
            return next((field for field in self._fields if field.name == item))
        
        # If item is list, return whose field's matches with item
        elif isinstance(item, list):
            return next((field for field in self._fields if set(field.fields) == set(item)))
        
        raise KeyError("Slicing must be either a `slice` or an `int` object.")
    
    def __add_fieldsetlist(self, fieldset: FieldsetList) -> "Sections":
        # If order in use, move it to end
        if (fs_order := fieldset.order) not in self._fields:
            order = fs_order
        else:
            # Get last order number and add one to create a new unique order
            order = max([order for order in self._fields.keys()]) + 1
        
        # Check name duplication
        if self.get(fieldset.name):
            raise KeyError("Name already in use!")
        
        self._fields[order] = fieldset
        return self
    
    # noinspection SpellCheckingInspection
    def add_fieldsetlist(self, fieldsetlist: FieldsetList) -> "Sections":
        """
            Adds a FieldsetList to this Fieldset's fields.
        :param fieldsetlist: The FieldsetList which should be added
        :return: self
        """
        # If fieldset already exists, integrate
        if fieldset := self.get(fieldsetlist.name):
            fieldset.integrate(fieldsetlist)
        else:
            self.__add_fieldsetlist(fieldsetlist)
        
        return self
    
    def _add_dict_fields(self, fields: dict):
        for name, nknown in fields.items():
            try:
                # Get fieldset
                fieldset = self[name]  # type: FieldsetList
            except StopIteration as e:
                if self.move_fields:
                    fieldset = self[0]
                else:
                    raise e
            
            # Add fields
            if isinstance(nknown, dict):
                fieldset.add_fields(
                    fields=nknown["fields"],
                    insert_prefix=nknown["name"]
                )
            elif isinstance(nknown, list):
                fieldset.add_fields(nknown)
            else:
                raise ValueError("The value of the dict must be a dict or a list!")
        
        return self
    
    def add_fields(self, fields: Union[list, dict]):
        """
            Adds fields to the FieldsetLists.
        :param fields: A dict containing as the key the name of the FieldSetList and as value the list with fields.
        The value can also be a dict, but must follow this scheme:
            fields: List, containing fields that should be added.
            name: str, this is the name for the insert_string_prefix.
        If list, it must contain a dict applying the rules above.
        :return: self
        """
        
        if isinstance(fields, dict):
            self._add_dict_fields(fields)
        elif isinstance(fields, list):
            for dct in fields:  # type: dict
                self._add_dict_fields(dct)
        
        return self
    
    def remove_fields(self, other: FieldsetList):
        """
            Removes a FieldsetList from this Fieldset.
        :param other: The FieldsetList to be removed.
        :return: self
        """
        del self._fields[other]
        
        return self
    
    def clear(self):
        """
            Calls the `clear_fields_copy` method of all FieldsetLists and updates them.
        :return: self
        """
        for field in self._fields:  # type: FieldsetList
            field.clear_fields_copy()
        
        return self
    
    def render(self, clear_before_rendering: bool = True, /) -> list:
        """
            Returns the FieldsetLists to a fieldset-like list, usable for Django.
        :return: List, usable as fieldset for Django
        """
        if clear_before_rendering:
            self.clear()
        
        return [
            sets.render()
            for sets in self._fields  # type: FieldsetList
        ]
    
    @property
    def fieldsets(self):
        for fieldset in self._fields:  # type: FieldsetList
            yield fieldset
    
    @property
    def fields(self) -> list:
        """
            Returns a copy of all FieldsetList's fields from this Fieldset.
        :return: List containing all fields of all FieldsetLists of this Fieldset
        """
        sets: FieldsetList
        return [sets.fields for sets in self._fields.values()]
    
    @property
    def flatten_fields(self) -> list:
        """
            Returns a copy of all FieldsetList's fields flatten from this Fieldset.
        :return: list containing all fields flatten of all FieldsetLists of this Fieldset
        """
        sets: FieldsetList
        return flat_list([sets.flatten_fields for sets in self._fields.values()])
    
    @staticmethod
    def from_fieldset(fieldset: list) -> "Sections":
        """
            Creates a Fieldset class from a Django fieldset.
        :param fieldset: The fieldset from django
        :return: A Fieldset
        """
        prepared = []
        
        for sets in fieldset:  # type: tuple
            appearance = sets[0]
            data = sets[1]
            
            prepared.append({
                "appearance": appearance,
                **data
            })
        
        return Sections(prepared)
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {' '.join([str(fs) for fs in self.fieldsets])}>"



from typing import *

from django.template.defaulttags import register

from ..libraries.typings import ModelInstance
from ..libraries.utils.text import (
    create_short, field_verbose, field_verbose_plural, listify, model_verbose, model_verbose_plural,
    textify,
)


@register.filter(name="listify")
def func_listify(x: List[str]) -> str:
    return listify(x)


@register.filter(name="textify")
def func_textify(html: str) -> str:
    return textify(html)


@register.filter(name="model_verbose")
def func_model_verbose(instance: ModelInstance) -> str:
    return model_verbose(instance)


@register.filter(name="model_verbose_plural")
def func_model_verbose_plural(instance: ModelInstance) -> str:
    return model_verbose_plural(instance)


@register.filter(name="field_verbose")
def func_field_verbose(instance: ModelInstance, field: str) -> str:
    return field_verbose(instance, field)


@register.filter(name="field_verbose_plural")
def func_field_verbose_plural(instance: ModelInstance, field: str) -> str:
    return field_verbose_plural(instance, field)


@register.simple_tag(name="format")
def func_format(string: str, *args, **kwargs) -> str:
    return string.format(*args, **kwargs)


@register.filter(name="create_short")
def func_create_short_filter(string: str, prefix: str) -> str:
    return create_short(string, prefix=prefix)


@register.simple_tag(name="create_short")
def func_create_short_simple_tag(*args, **kwargs) -> str:
    return create_short(*args, **kwargs)



from setuptools import setup, find_packages


long_description = "This package provides you some utils I think are useful. The package follows the " \
                       "modular-principle, instead of rewriting fields, admin pages etc., you create a mixin and " \
                       "reuse it for your projects!\n\nPlease see the package homepage for the documentation."

current_version = "1.1.1"

setup(
    name="django-common-utils",
    packages=find_packages(),
    version=str(current_version),
    license="MIT",
    description="This package provides you some utils I think are useful.",
    long_description=long_description,
    author="Myzel394",
    author_email="myzel394.xyllian@gmail.com",
    url="https://github.com/Myzel394/django-common-utils",
    download_url="https://github.com/Myzel394/django-common-utils/archieve/v_" + str(current_version) + ".tar.gz",
    keywords=["django", "django-utils", "django-common", "django-models", "django-admin"],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Build Tools"
    ],
    
)



from typing import KeysView, ValuesView

from django.template.defaulttags import register


@register.simple_tag
def as_list(*args) -> list:
    return [*args]


@register.simple_tag
def as_set(*args) -> set:
    return set(args)


@register.simple_tag
def as_dict(**kwargs) -> dict:
    return kwargs


@register.filter
def as_list(value) -> list:
    return list(value)


@register.filter
def as_set(value) -> set:
    return set(value)


@register.simple_tag
@register.filter
def get_keys(value: dict) -> KeysView:
    return value.keys()


@register.simple_tag
@register.filter
def get_values(value: dict) -> ValuesView:
    return value.values()



from types import FunctionType, ModuleType
from typing import *


def create_func(func: FunctionType) -> FunctionType:
    """Creates the actual function"""
    
    def actual_func(*args, **kwargs):
        return func(*args, **kwargs)
    
    return actual_func


def create_name(module: ModuleType, func_name: str) -> str:
    """Creates the name for the tag"""
    
    return f"{module.__name__}.{func_name}"


def iter_module(module: ModuleType) -> Generator[Tuple[str, FunctionType], ModuleType, None]:
    """Iterates over the module's functions"""
    
    for name in dir(module):
        
        if not (name.startswith("_") or name.endswith("_")) and \
                hasattr((func := getattr(module, name)), "__call__"):
            yield create_name(module, name), create_func(func)



import re
import unicodedata
from abc import abstractmethod
from typing import Optional

from django.core.exceptions import ValidationError
from django.core.validators import validate_slug
from django.db import models
from django.urls import reverse
from django.utils.text import slugify
from django.utils.translation import gettext as _

from ....constants import RFC
from ..get_settings import extract_model_kwargs as ek


class UrlBaseMixin:
    @staticmethod
    @abstractmethod
    def get_url_field() -> str:
        raise NotImplementedError("Method is not implemented.")
    
    @staticmethod
    @abstractmethod
    def get_url() -> str:
        """Returns the string to reverse the url"""
        raise NotImplementedError("Method is not implemented.")
    
    @property
    def url_value(self):
        return getattr(self, self.get_url_field())
    
    @property
    def reversed_url(self) -> str:
        return reverse(self.get_url(), args=(self.url_value,))


class SlugMixin(models.Model):
    """Adds a `slug` field, which value will automatically be created from another field. Specify the field using
    `SLUG_TARGETED_FIELD`"""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    ___COMMON_SLUG_CHANGE_THRESHOLD = 24
    
    slug = models.SlugField(
        **ek(___common_name, "slug", {
            "verbose_name": _("Slug"),
            "help_text": _("Representation of the object in the url. Can be left out blank"),
            "max_length": 127,
            "validators": [validate_slug],
            "unique": True,
            "blank": True,
        })
    )  # type: str
    
    def save(self, *args, **kwargs):
        used_slugs = set(
            self.__class__
                .objects
                .all()
                .values_list("slug", flat=True)
        )
        
        if self.slug is None or self.slug == "":
            counter: Optional[int] = None
            
            while True:
                use_slug = self.slugify(getattr(self, self.___COMMON_SLUG_TARGETED_FIELD()))
                
                if type(counter) is int:
                    # Appending counter and increasing it
                    use_slug += str(counter)
                    counter += 1
                else:
                    # Preparing counter
                    counter = 0
                
                if use_slug not in used_slugs:
                    break
            
            self.slug = use_slug
        else:
            if self.slug in used_slugs:
                raise ValidationError(_("Slug is already used!"))
        
        return super().save(*args, **kwargs)
    
    @staticmethod
    @abstractmethod
    def ___COMMON_SLUG_TARGETED_FIELD() -> str:
        raise NotImplementedError(f"{__qualname__} is not implemented")
    
    @staticmethod
    def get_url_field() -> str:
        return "slug"
    
    @staticmethod
    def slugify(slug: str) -> str:
        use_slug: str = slugify(slug)
        # Resolve special characters
        use_slug = unicodedata \
            .normalize('NFKD', use_slug) \
            .encode("ascii", "ignore") \
            .decode("ascii")
        # Remove not resolved special characters
        use_slug = re.sub(rf"[^{RFC.URL_ALLOWED_CHARS_3986}]", "", use_slug.lower())
        
        return use_slug



from ..typings import Kwargs
from ..utils import EnsureIterationDictType

__all__ = [
    "HandleOn", "TextOptimizerDefault", "HTMLOptimizerDefault",
    "AddAttributesDict", "UnwrapDict"
]

AddAttributesDict = EnsureIterationDictType[
    str, EnsureIterationDictType[str, str]
]

UnwrapDict = EnsureIterationDictType[str, str]


class HandleOn:
    SAVE = "SAVE"
    CREATION = "CREATION"
    DELETION = "DELETION"


class TextOptimizerDefault:
    space_after: str = r".,!?:;)+*&§%|\/\\"
    space_before: str = r"\(+*&%#$|\/\\"
    no_space_before: str = "?!.,`':;-"
    no_space_after: str = "-#"
    ignore_for_digits: str = ".,"


class HTMLOptimizerDefault:
    unwrap: UnwrapDict = {
        ("p",): {"img", "table"}
    }
    add_attributes: AddAttributesDict = {
        "img": {
            "loading": "lazy"
        },
        "a": {
            "rel": "noopener noreferrer"
        }
    }
    minify_opts: Kwargs = {
        "remove_comments": True,
        "remove_empty_space": True,
        "remove_all_empty_space": False,
        "reduce_empty_attributes": True,
        "reduce_boolean_attributes": True,
        "remove_optional_attribute_quotes": True,
        "convert_charrefs": True
    }



from django.utils.translation import gettext as _

from ...utils.settings import get_setting
from ..admin import AdminMixinsMixin
from ..sections import FieldsetList, Sections

__all__ = [
    "DefaultAdminMixin"
]


class DefaultAdminMixin(AdminMixinsMixin):
    def get_section(self) -> Sections:
        return Sections([
            FieldsetList(
                order=0,
                appearance=None,
                name="default",
                fields=[]
            ),
            FieldsetList(
                order=1,
                appearance=get_setting("ADMIN_EXTRA_SETTINGS", _("Special settings")),
                name="extra",
                fields=[]
            ),
            FieldsetList(
                order=2,
                appearance=get_setting("ADMIN_ADVANCED_SETTINGS", _("Advanced settings")),
                classes=["collapse", ],
                name="advanced",
                description=_("Advanced settings. Change them with caution!"),
                fields=[]
            ),
            FieldsetList(
                order=3,
                appearance=get_setting("ADMIN_ON_CREATED_FIELD", _("After creation")),
                classes=["collapse", ],
                name="created",
                description=_("Settings visible after creation of the object"),
                fields=[]
            )
        ])
    
    fieldset_fields = {
        "default": ["...!"],
        "extra": ["...!"],
        "advanced": ["...!"],
        "created": ["id", "...!"]
    }



from django.conf import settings
from django.db import models
from django.utils.translation import gettext as _
from django_hint import QueryType

from ...utils.text import model_verbose
from ..get_settings import extract_model_kwargs as ek


class AutomaticUserAssociationCreationMixin(models.Model):
    """Automatically associates the user who created this object to the `added_by` field. Must be enabled in admin"""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    
    added_by = models.ForeignKey(
        **ek(___common_name, "added_by", {
            "to": settings.AUTH_USER_MODEL,
            "verbose_name": model_verbose(settings.AUTH_USER_MODEL),
            "editable": False,
            "on_delete": models.CASCADE,
        }),
    )  # type: settings.AUTH_USER_MODEL


class AuthorMixin(models.Model):
    """Adds an `author` field."""
    
    class Meta:
        abstract = True
    
    ___common_name = __qualname__
    ___COMMON_USER_MODEL = settings.AUTH_USER_MODEL
    
    authors = models.ManyToManyField(
        **ek(___common_name, "authors", {
            "to": settings.AUTH_USER_MODEL,
            "verbose_name": _("Authors"),
        })
    )  # type: QueryType[settings.AUTH_USER_MODEL]
    
    def is_user_author(self, user) -> bool:
        return self.authors.only("id").filter(id=user.id).exists()



import logging
from collections import Iterable as collections_Iterable
from typing import *

from django.contrib import admin
from django.core.exceptions import FieldDoesNotExist
from django.db.models.options import Options
from django.http import HttpRequest
from django.utils.translation import gettext as _

from .sections import FieldsetList, Sections
from ..typings import *
from ..utils import camelcase_to_underscore, model_verbose


class AdminFieldsetMixin:
    name = "__"
    
    def __init__(self):
        self.name = camelcase_to_underscore(
            self.__class__.__name__
                .replace("Admin", "")
                .replace("Fieldset", "")
                .replace("Mixin", "")
        )
    
    def get_mixin_fields(self, **_) -> Dict[str, List[str]]:
        return {}
    
    def get_fieldsetslist(self, **_) -> Union[list, FieldsetList, None]:
        return None
    
    def get_readonly_fields(self, **_) -> Union[Iterable[str], str, None]:
        return None
    
    def get_search_fields(self, **_) -> Union[list, str, set, None]:
        return None
    
    def get_list_display(self, **_) -> Union[list, str, set, None]:
        return None
    
    def get_list_display_links(self, **_) -> Union[list, str, set, None]:
        return None
    
    def get_list_filter(self, **_) -> Union[list, str, set, None]:
        return None
    
    def get_autocomplete_fields(self, **_) -> Union[list, str, set, None]:
        return None
    
    def get_actions(self, **_) -> Union[list, str, set, None]:
        return None


class BaseAdminMixinsMixin:
    mixins = []
    order_by_required = True  # type: Optional[bool]
    fieldset_fields = {
        "default": ["...!"],
        "advanced": ["...!"],
        "created": ["...!"]
    }
    fieldset_descriptions = {}
    make_not_editable_readonly = True
    allow_mixin_methods = "*"  # type: Union[list, str]
    
    def get_section(self) -> Sections:
        # Default sections
        return Sections([
            FieldsetList(
                order=0,
                appearance=None,
                name="default",
                fields=[]
            ),
            FieldsetList(
                order=2,
                appearance=_("Advanced Settings"),
                classes=["collapse", ],
                name="advanced",
                description=_("Pay attention when editing these fields."),
                fields=[]
            ),
            FieldsetList(
                order=3,
                appearance=_("Visible after creation"),
                classes=["collapse", ],
                name="created",
                description=_("These fields are visible after this object got created."),
                fields=[]
            )
        ])
    
    def get_collected_fieldsets(self) -> Sections:
        fieldset = self.get_section()
        
        for fs in self.get_fieldsetslist():  # type: FieldsetList
            fieldset.add_fieldsetlist(fs)
        
        return fieldset
    
    def __get_fields(self, method: str, **kwargs) -> list:
        """
            Collects all fields from this class using a given method.
            Return types of the method should be a str, Iterable or None.
            If returned type is an Iterable, these fields will be added.
            If returned type is None, no field will be added.
            If returned type is anything else, it'll added to a list and that list will be added.
        :param method: The method that should be executed
        :param kwargs: **kwargs that should be passed to the method
        :return: set containing all fields
        """
        if self.allow_mixin_methods != "*" and (isinstance(method, list) and method not in self.allow_mixin_methods):
            raise ValueError(f"Method \"{method}\" is not allowed for mixins!")
        
        all_fields = []
        
        for mixin in self.mixins:  # type: AdminFieldsetMixin.__class__
            init = mixin()  # type: AdminFieldsetMixin
            fields = getattr(init, method)(**kwargs)
            
            # Add fields
            # Append, if it's a str
            if isinstance(fields, collections_Iterable) and type(fields) is not str:
                all_fields.extend(fields)
            elif fields is not None:
                all_fields.append(fields)
        
        # Add super() fields, if super() has this method
        if hasattr(super(), method):
            all_fields += getattr(super(), method)(**kwargs)
        
        return all_fields
    
    def __get_fieldsets(self, **kwargs) -> Sections:
        fieldset = self.get_collected_fieldsets()
        
        # Add fieldsets of mixins
        for fieldsetlist in self.get_fieldsetslist(**kwargs):  # type: FieldsetList
            fieldset.add_fieldsetlist(fieldsetlist)
        fieldset.add_fields(self.fieldset_fields)  # Add given fields from ModelAdmin
        
        # Add fields of mixins
        fieldset.add_fields(self.get_mixin_fields(**kwargs))
        
        for name, description in self.fieldset_descriptions.items():  # type: str
            if description not in getattr((fl := fieldset[name]), "description"):
                fl.description += "\n" + description
        
        return fieldset
    
    @staticmethod
    def __get_kwargs(*args) -> dict:
        request, obj = args
        
        if obj is HttpRequest:
            obj, request = request, obj
        
        return {
            "obj": obj,
            "request": request
        }
    
    def get_fieldsets(self, request=None, obj=None) -> list:
        return self.__get_fieldsets(**self.__get_kwargs(request, obj)).render()
    
    def get_readonly_fields(self, request=None, obj=None) -> list:
        # Get predefined readonly fields and from mixins
        readonly = \
            list(super().get_readonly_fields(request, obj)) + \
            self.__get_fields("get_readonly_fields", **self.__get_kwargs(request, obj))
        
        if self.make_not_editable_readonly:
            # noinspection PyProtectedMember
            meta = self.model._meta  # type: Options
            
            for field in self.__get_fieldsets(request=request, obj=obj).flatten_fields:  # type: str
                try:
                    # If field is not editable, add it to the readonly fields
                    if not meta.get_field(field).editable:
                        readonly.append(field)
                # If field does not exist, just continue. Maybe it`s a custom property from the admin class
                except FieldDoesNotExist:
                    continue
        
        return readonly
    
    def get_search_fields(self, request):
        return self.__get_fields("get_search_fields", request=request)
    
    def get_list_display(self, request):
        return self.__get_fields("get_list_display", request=request)
    
    def get_list_display_links(self, request, list_display):
        return self.__get_fields("get_list_display_links", request=request, list_display=list_display)
    
    def get_list_filter(self, request):
        return self.__get_fields("get_list_filter", request=request)
    
    def get_autocomplete_fields(self, request):
        return self.__get_fields("get_autocomplete_fields", request=request)
    
    # <-> Helpers <-> #
    
    def get_true_mixins(self) -> Generator[Type[AdminFieldsetMixin], None, None]:
        for mixin in self.mixins:
            if issubclass(mixin, AdminFieldsetMixin):
                yield mixin
    
    def get_mixin_fields(self, **kwargs) -> list:
        """
            Collects all fields from the mixins.
        :param kwargs: kwargs to be passed to the `get_mixin_fields` method
        :return: dict containing all fields as value and name as key
        """
        all_fields = []
        
        name: str
        value: list
        
        # Iterate through all fields from the mixins
        for mixin in self.get_true_mixins():
            instance = mixin()  # type: AdminFieldsetMixin
            
            # Get fields
            for name, fields in instance.get_mixin_fields(**kwargs).items():
                all_fields.append({
                    name: {
                        "fields": fields,
                        "name": instance.name
                    }
                })
        
        return all_fields
    
    def get_fieldsetslist(self, **kwargs) -> list:
        return self.__get_fields("get_fieldsetslist", **kwargs)
    
    def get_view_on_site_url(self, obj: Optional[ModelInstance] = None) -> Optional[str]:
        if obj:
            try:
                obj.get_url()
            except NotImplementedError:
                logging.warning(f"You forgot to implement some methods on `{model_verbose(obj)}`")
            else:
                return obj.reversed_url
        return


class AdminMixinsMixin(BaseAdminMixinsMixin, admin.ModelAdmin):
    pass



from datetime import datetime, timedelta

from ..admin import AdminFieldsetMixin


class SlugAdminFieldsetMixin(AdminFieldsetMixin):
    def get_readonly_fields(self, obj=None, **_):
        if obj:
            # If the object has a pub_date, make slug only readonly if the obj's pub_date is older than the given
            # threshold.
            if (date := getattr(obj, "pub_date")) or (date := getattr(obj, "edited_at")):
                now = datetime.now()
                if date > now - timedelta(hours=obj.___COMMON_SLUG_CHANGE_THRESHOLD):
                    return
            return "slug"
    
    def get_mixin_fields(self, **_) -> dict:
        return {
            "default": ["slug", ]
        }



from typing import *

__all__ = [
    "ensure_iteration", "ensure_dict"
]


# https://stackoverflow.com/a/3655857/9878135
def islambda(x, /) -> bool:
    lambda_func = lambda: 0
    return isinstance(x, type(lambda_func)) and x.__name__ == lambda_func.__name__


def ensure_iteration(value, targeted_type) -> Generator[Any, Any, None]:
    """Iterates over `value` if it is not type of `targeted_value`. Otherwise `value` will be yield directly.
    Basically takes care if there are values in a iterable or if the value is passed solo."""
    
    if islambda(targeted_type):
        if targeted_type(value):
            yield value
    else:
        if type(value) is targeted_type:
            yield value
    
    for val in value:
        yield val


def ensure_dict(value: dict, key_type, value_type) -> Generator[Tuple[Any, Any], Any, None]:
    """Basically like `ensure_iteration` for key and value. For each key each value will be returned"""
    
    for key_unknown, value_unknown in value.items():
        for key in ensure_iteration(key_unknown, key_type):
            for value in ensure_iteration(value_unknown, value_type):
                yield key, value
